"""File gRPC client."""
from typing import Any, Dict, Generator, Iterable, Iterator, cast
from uuid import UUID

import grpc
from fred_api.fileman import service_fileman_grpc_pb2_grpc
from fred_api.fileman.service_fileman_grpc_pb2 import CreateRequest, ReadRequest, StatRequest
from frgal import GrpcClient, GrpcDecoder


class FileClient(GrpcClient):
    """File gRPC client."""

    grpc_service = 'File'

    def __init__(self, netloc: str, credentials: grpc.ChannelCredentials = None):
        """Initialize File client.

        Arguments:
            netloc: Network location of a gRPC server.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.

        Raises:
            FileNotFoundError: If the file is not found.
            ConnectionAbortedError: Another error in connection occured.
        """
        decoder = GrpcDecoder()
        super().__init__(netloc, decoder=decoder, grpc_modules=[service_fileman_grpc_pb2_grpc], credentials=credentials)

    def stat(self, uuid: UUID) -> Dict[str, Any]:
        """Get file metadata.

        Arguments:
            uuid: A file UUID.
        """
        request = StatRequest()
        request.uuid.value = str(uuid)
        try:
            return cast(Dict[str, Any], self.call(self.grpc_service, 'stat', request))
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.NOT_FOUND:
                raise FileNotFoundError('File {} not found.'.format(uuid))
            raise ConnectionAbortedError('Connection failed: {}'.format(error))

    def read(self, uuid: UUID, size: int = 0) -> Generator[bytes, None, None]:
        """Get file content.

        Arguments:
            uuid: A file UUID.
            size: A maximum size of a chunk.

        Raises:
            FileNotFoundError: If the file is not found.
            ConnectionAbortedError: Another error in connection occured.
        """
        request = ReadRequest()
        request.uuid.value = str(uuid)
        request.size = size
        try:
            return cast(Generator[bytes, None, None], self.call_stream(self.grpc_service, 'read', request))
        except grpc.RpcError as error:
            if error.code() == grpc.StatusCode.NOT_FOUND:
                raise FileNotFoundError('File {} not found.'.format(uuid))
            raise ConnectionAbortedError('Connection failed: {}'.format(error))

    def _make_create_requests(self, name: str, data: Iterable[bytes], mimetype: str = None) -> Iterator[CreateRequest]:
        """Return an iterator of create requests."""
        # Turn data into an iterator.
        for i, chunk in enumerate(data):
            request = CreateRequest()
            request.data = chunk
            if i == 0:
                request.name = name
                if mimetype:
                    request.mimetype = mimetype
            yield request

    def create(self, name: str, data: Iterable[bytes], *, mimetype: str = None) -> UUID:
        """Create a new file.

        Arguments:
            name: A file name.
            data: A file content.
            mimetype: A file mimetype.

        Raises:
            ConnectionAbortedError: An error in connection occured.
        """
        try:
            result = self.call(self.grpc_service, 'create', self._make_create_requests(name, data, mimetype))
        except grpc.RpcError as error:
            raise ConnectionAbortedError('Connection failed: {}'.format(error))
        return UUID(result)
